class CumulusCIException(Exception):
    """Base class for all CumulusCI Exceptions"""

    pass


class CumulusCIUsageError(CumulusCIException):
    """An exception thrown due to improper usage which should be resolvable by proper usage"""

    pass


class CumulusCIFailure(CumulusCIException):
    """An exception representing a failure such as a Metadata deployment failure or a test failure.  CI systems can handle these to determine fail vs error status"""

    pass


class NotInProject(CumulusCIUsageError):
    """Raised when no project can be found in the current context"""

    pass


class ProjectConfigNotFound(CumulusCIUsageError):
    """Raised when a project is found in the current context but no configuration was found for the project"""

    pass


class ProjectMinimumVersionError(CumulusCIException):
    pass


class KeychainNotFound(CumulusCIException):
    """Raised when no keychain could be found"""

    pass


class KeychainKeyNotFound(CumulusCIUsageError):
    """Raised when the keychain key couldn't be found"""

    pass


class OrgNotFound(CumulusCIUsageError):
    """Raised when no org could be found by a given name in the project keychain"""

    pass


class OrgCannotBeLoaded(CumulusCIUsageError):
    """Raised when an org cannot be decrypted or unpickled"""

    pass


class ServiceNotConfigured(CumulusCIUsageError):
    """Raised when no service configuration could be found by a given name in the project keychain"""

    pass


class ServiceCannotBeLoaded(CumulusCIUsageError):
    """Raised when a service cannot be decrypted or unpickled"""

    pass


class ServiceNotValid(CumulusCIUsageError):
    """Raised when no service configuration could be found by a given name in the project configuration"""

    pass


class DependencyResolutionError(CumulusCIException):
    """Raised when an issue is encountered while resolving a static dependency map"""

    pass


class DependencyParseError(CumulusCIException):
    """Raised when CumulusCI cannot parse a dependency specification."""

    pass


class ConfigError(CumulusCIException):
    """Raised when a configuration enounters an error"""

    def __init__(self, message=None, config_name=None):
        super(ConfigError, self).__init__(message)
        self.message = message
        self.config_name = config_name

    def __str__(self):
        return f"{self.message} for config {self.config_name}"


class ConfigMergeError(ConfigError):
    """Raised when merging configuration fails."""

    pass


class AntTargetException(CumulusCIException):
    """Raised when a generic Ant target error occurs"""

    pass


class DeploymentException(CumulusCIFailure):
    """Raised when a metadata api deployment error occurs"""

    pass


class ApexTestException(CumulusCIFailure):
    """Raised when a build fails because of an Apex test failure"""

    pass


class SalesforceCredentialsException(CumulusCIException):
    """Raise when Salesforce credentials are invalid"""

    pass


class TaskRequiresSalesforceOrg(CumulusCIUsageError):
    """Raise when a task that requires a Salesforce org_config is not initialized with an org_config"""

    pass


class TaskOptionsError(CumulusCIUsageError):
    """Raise when a task's options are invalid"""

    pass


class TaskImportError(CumulusCIException):
    """Raise when a task's class can't be loaded"""

    pass


class NamespaceNotFoundError(CumulusCIUsageError):
    """Raise when namespace is not found in project includes"""


class TaskNotFoundError(CumulusCIUsageError):
    """Raise when task is not found in project config"""

    pass


class FlowInfiniteLoopError(CumulusCIUsageError):
    """Raised when a flow configuration creates a infinite loop"""

    pass


class FlowConfigError(CumulusCIException):
    """Raised when a flow configuration encounters an error"""

    pass


class FlowNotFoundError(CumulusCIUsageError):
    """Raise when flow is not found in project config"""

    pass


class FlowNotReadyError(CumulusCIException):
    """Raise when flow is called before it has been prepared"""

    pass


class SfdxOrgException(CumulusCIException):
    """Raise for errors related to fetching orgs from the sfdx keychain"""


class ScratchOrgException(SfdxOrgException):
    """Raise for errors related to scratch orgs"""

    pass


class GithubException(CumulusCIException):
    """Raise for errors related to GitHub"""

    pass


class GithubApiError(CumulusCIException):
    pass


class GithubApiNotFoundError(CumulusCIException):
    pass


class GithubApiNoResultsError(CumulusCIException):
    pass


class GithubApiUnauthorized(CumulusCIException):
    pass


class SalesforceException(CumulusCIException):
    """Raise for errors related to Salesforce"""

    pass


class SalesforceDXException(CumulusCIException):
    """Raise for errors related to Salesforce DX"""

    pass


class SOQLQueryException(CumulusCIFailure):
    """Raise for errors related to Salesforce DX"""

    pass


class CommandException(CumulusCIFailure):
    """Raise for errors coming from spawned CLI subprocesses"""

    pass


class BrowserTestFailure(CumulusCIFailure):
    """Raise when browser tests fail"""

    pass


class ApexCompilationException(CumulusCIFailure):
    """Raise when apex compilation fails"""

    def __str__(self):
        line, problem = self.args
        return f"Apex compilation failed on line {line}: {problem}"


class ApexException(CumulusCIFailure):
    """Raise when an Apex Exception is raised in an org"""

    def __str__(self):
        message, stacktrace = self.args
        stacktrace = "\n  ".join(stacktrace.splitlines())
        return f"Apex error: {message}\n  Stacktrace:\n  {stacktrace}"


class PushApiObjectNotFound(CumulusCIException):
    """Raise when Salesforce Push API object is not found"""

    pass


class RobotTestFailure(CumulusCIFailure):
    """Raise when a robot test fails in a test suite"""

    pass


class BulkDataException(CumulusCIFailure):
    """Raise for errors from bulkdata tasks"""

    pass


class DependencyLookupError(CumulusCIFailure):
    """Raise for errors looking up the PackageVersionId for a managed package dependency"""

    pass


class PackageUploadFailure(CumulusCIFailure):
    """Raise for errors in uploading a 2GP package version"""

    pass


class PackageInstallError(Exception):
    """Error installing a 2GP package version"""


class YAMLParseException(CumulusCIException):
    """Error parsing a YAML File"""



from unittest import mock

from cumulusci.core.config import ServiceConfig, TaskConfig
from cumulusci.tasks.github import PullRequests
from cumulusci.tests.util import create_project_config


class TestPullRequests:
    def test_run_task(self):
        project_config = create_project_config()
        project_config.keychain.set_service(
            "github",
            "test_alias",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "token": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )
        task_config = TaskConfig()
        task = PullRequests(project_config, task_config)
        repo = mock.Mock()
        repo.pull_requests.return_value = [mock.Mock(number=1, title="Test PR")]
        task.get_repo = mock.Mock(return_value=repo)
        task.logger = mock.Mock()
        task()
        task.logger.info.assert_called_with("#1: Test PR")



from cumulusci.tests.pytest_plugins.vcr_string_compressor import (
    StringToTemplateCompressor,
)

tpl_dir = "cumulusci/tests/shared_cassettes/vcr_string_templates"


example_string = """
<?xml version="1.0" encoding="UTF-8"?><jobInfo
   xmlns="http://www.force.com/2009/06/asyncapi/dataload">
 <id>75023000003RqO6AAK</id>
 <operation>insert</operation>
 <object>Account</object>
 <createdById>00523000002KbrQAAS</createdById>
 <createdDate>2022-03-29T20:03:56.000Z</createdDate>
 <systemModstamp>2022-03-29T20:03:56.000Z</systemModstamp>
 <state>Open</state>
 <concurrencyMode>Parallel</concurrencyMode>
 <contentType>CSV</contentType>
 <numberBatchesQueued>0</numberBatchesQueued>
 <numberBatchesInProgress>0</numberBatchesInProgress>
 <numberBatchesCompleted>0</numberBatchesCompleted>
 <numberBatchesFailed>0</numberBatchesFailed>
 <numberBatchesTotal>0</numberBatchesTotal>
 <numberRecordsProcessed>0</numberRecordsProcessed>
 <numberRetries>0</numberRetries>
 <apiVersion>vxx.0</apiVersion>
 <numberRecordsFailed>0</numberRecordsFailed>
 <totalProcessingTime>0</totalProcessingTime>
 <apiActiveProcessingTime>0</apiActiveProcessingTime>
 <apexProcessingTime>0</apexProcessingTime>
</jobInfo>""".strip()


class TestVcrStringToTemplateCompressor:
    def test_simple_string_compression(self):
        sc = StringToTemplateCompressor.from_directory(tpl_dir)
        data = sc.string_to_template_if_possible(example_string)
        assert data == {
            "include_template": "jobInfo_insert_xml.tpl",
            "vars": {
                "id": "75023000003RqO6AAK",
                "creator": "00523000002KbrQAAS",
                "cdate": "2022-03-29T20:03:56.000Z",
                "smts": "2022-03-29T20:03:56.000Z",
                "qdbatches": "0",
                "numbatchtotal": "0",
            },
        }



from unittest import mock

import pytest

from cumulusci.utils.waiting import poll, retry


def test_retry(caplog):
    func = mock.Mock(side_effect=[Exception, 42])
    assert retry(func) == 42
    assert "Sleeping for 5 seconds before retry..." in caplog.text
    assert "Retrying (4 attempts remaining)" in caplog.text


def test_retry__limit():
    func = mock.Mock(side_effect=Exception)
    with pytest.raises(Exception):
        retry(func, retries=0)
    assert func.call_count == 1


def test_poll():
    func = mock.Mock(side_effect=[False, False, False, True])
    poll(func)
    assert func.call_count == 4



import json
import os
import unittest
from unittest.mock import Mock, call, mock_open, patch

from cumulusci.tasks.salesforce.salesforce_files import (
    ListFiles,
    RetrieveFiles,
    UploadFiles,
)
from cumulusci.tasks.salesforce.tests.util import create_task


class TestDisplayFiles:
    def test_display_files(self):
        task = create_task(ListFiles, {})
        task._init_api = Mock()
        task._init_api.return_value.query.return_value = {
            "totalSize": 2,
            "records": [
                {"Title": "TEST1", "Id": "0PS000000000000", "FileType": "TXT"},
                {"Title": "TEST2", "Id": "0PS000000000001", "FileType": "TXT"},
            ],
        }
        task()

        task._init_api.return_value.query.assert_called_once_with(
            "SELECT Title, Id, FileType FROM ContentDocument"
        )
        assert task.return_values == [
            {"Id": "0PS000000000000", "FileName": "TEST1", "FileType": "TXT"},
            {"Id": "0PS000000000001", "FileName": "TEST2", "FileType": "TXT"},
        ]


class TestRetrieveFiles(unittest.TestCase):
    @patch("requests.get")
    @patch("os.path.exists")
    @patch("os.makedirs")
    @patch("builtins.open")
    def test_run_task(self, mock_open, mock_makedirs, mock_exists, mock_get):
        # Mock Salesforce query response
        mock_sf = Mock()
        mock_sf.query.return_value = {
            "totalSize": 2,
            "records": [
                {
                    "Title": "TEST1",
                    "Id": "0PS000000000000",
                    "FileType": "TXT",
                    "VersionData": "version1",
                    "ContentDocumentId": "doc1",
                },
                {
                    "Title": "TEST2",
                    "Id": "0PS000000000001",
                    "FileType": "TXT",
                    "VersionData": "version2",
                    "ContentDocumentId": "doc2",
                },
            ],
        }

        # Mock org config
        mock_org_config = Mock()
        mock_org_config.instance_url = "https://test.salesforce.com"
        mock_org_config.access_token = "test token"

        # Create task with mocked Salesforce and org config
        task = create_task(RetrieveFiles, {"path": "test_dir", "file_list": ""})
        task.sf = mock_sf
        task.org_config = mock_org_config

        # Mock file existence and request response
        mock_exists.return_value = False
        mock_response = Mock()
        mock_response.iter_content.return_value = [b"chunk1", b"chunk2"]
        mock_response.raise_for_status = Mock()
        mock_get.return_value = mock_response

        # Run the task
        task._run_task()

        # Check if query was called with correct SOQL
        mock_sf.query.assert_called_once_with(
            "SELECT Title, Id, FileType, VersionData, ContentDocumentId FROM ContentVersion WHERE isLatest=true "
        )

        # Check if files are downloaded
        expected_calls = [
            call(
                "https://test.salesforce.com/version1",
                headers={"Authorization": "Bearer test token"},
                stream=True,
            ),
            call(
                "https://test.salesforce.com/version2",
                headers={"Authorization": "Bearer test token"},
                stream=True,
            ),
        ]
        mock_get.assert_has_calls(expected_calls, any_order=True)

        # Check if files are written correctly
        mock_open.assert_any_call(os.path.join("test_dir", "TEST1.txt"), "wb")
        mock_open.assert_any_call(os.path.join("test_dir", "TEST2.txt"), "wb")

        # Check if return values are set correctly
        self.assertEqual(
            task.return_values,
            [
                {
                    "Id": "0PS000000000000",
                    "FileName": "TEST1",
                    "FileType": "TXT",
                    "VersionData": "version1",
                    "ContentDocumentId": "doc1",
                },
                {
                    "Id": "0PS000000000001",
                    "FileName": "TEST2",
                    "FileType": "TXT",
                    "VersionData": "version2",
                    "ContentDocumentId": "doc2",
                },
            ],
        )


class TestUploadFiles(unittest.TestCase):
    @patch("requests.post")
    @patch("os.listdir")
    @patch("os.path.isfile")
    @patch("builtins.open", new_callable=mock_open, read_data=b"test data")
    def test_run_task(self, mock_open, mock_isfile, mock_listdir, mock_post):
        # Mock org config and project config
        mock_org_config = Mock()
        mock_org_config.instance_url = "https://test.salesforce.com"
        mock_org_config.access_token = "test token"

        mock_project_config = Mock()
        mock_project_config.project__package__api_version = "50.0"

        # Create task with mocked configs
        task = create_task(UploadFiles, {"path": "test_dir", "file_list": ""})
        task.org_config = mock_org_config
        task.project_config = mock_project_config

        # Mock file discovery
        mock_listdir.return_value = ["file1.txt", "file2.txt"]
        mock_isfile.side_effect = lambda filepath: filepath in [
            os.path.join("test_dir", "file1.txt"),
            os.path.join("test_dir", "file2.txt"),
        ]

        # Mock requests response
        mock_response = Mock()
        mock_response.status_code = 201
        mock_response.json.return_value = {"id": "contentversionid"}
        mock_post.return_value = mock_response

        # Run the task
        task._run_task()

        mock_open.assert_any_call(os.path.join("test_dir", "file1.txt"), "rb")
        mock_open.assert_any_call(os.path.join("test_dir", "file2.txt"), "rb")

        # Check if requests.post was called correctly
        expected_calls = [
            call(
                "https://test.salesforce.com/services/data/v50.0/sobjects/ContentVersion/",
                headers={"Authorization": "Bearer test token"},
                files={
                    "entity_content": (
                        "",
                        json.dumps(
                            {
                                "Title": "file1",
                                "PathOnClient": os.path.join("test_dir", "file1.txt"),
                            }
                        ),
                        "application/json",
                    ),
                    "VersionData": (
                        "file1.txt",
                        mock_open(),
                        "application/octet-stream",
                    ),
                },
            ),
            call(
                "https://test.salesforce.com/services/data/v50.0/sobjects/ContentVersion/",
                headers={"Authorization": "Bearer test token"},
                files={
                    "entity_content": (
                        "",
                        json.dumps(
                            {
                                "Title": "file2",
                                "PathOnClient": os.path.join("test_dir", "file2.txt"),
                            }
                        ),
                        "application/json",
                    ),
                    "VersionData": (
                        "file2.txt",
                        mock_open(),
                        "application/octet-stream",
                    ),
                },
            ),
        ]

        self.assertEqual(
            task.return_values,
            [
                {
                    "Title": "file1",
                    "PathOnClient": os.path.join("test_dir", "file1.txt"),
                },
                {
                    "Title": "file2",
                    "PathOnClient": os.path.join("test_dir", "file2.txt"),
                },
            ],
        )

        mock_post.assert_has_calls(expected_calls, any_order=True)



import sys
from abc import abstractmethod
from typing import Optional, Type

from cumulusci.core.config import BaseProjectConfig, UniversalConfig
from cumulusci.core.debug import DebugMode, get_debug_mode
from cumulusci.core.exceptions import NotInProject, ProjectConfigNotFound
from cumulusci.core.flowrunner import FlowCallback, FlowCoordinator
from cumulusci.core.keychain import BaseProjectKeychain


# pylint: disable=assignment-from-none
class BaseCumulusCI:
    universal_config_class = UniversalConfig
    project_config_class = BaseProjectConfig
    keychain_class = BaseProjectKeychain
    callback_class = FlowCallback

    universal_config: UniversalConfig
    project_config: Optional[BaseProjectConfig]
    keychain: Optional[BaseProjectKeychain]
    debug_mode: DebugMode
    project_config_error: Exception

    def __init__(self, *args, load_keychain=True, **kwargs):
        self.keychain = None
        self.debug_mode = get_debug_mode()

        self._load_universal_config()

        try:
            self._load_project_config(*args, **kwargs)
            self._add_repo_to_path()
        except (NotInProject, ProjectConfigNotFound) as e:
            self.project_config = None
            self.project_config_error = e
        if load_keychain:
            self._load_keychain()

    @property
    def universal_config_cls(self) -> Type:
        klass = self.get_universal_config_class()
        return klass or self.universal_config_class

    @abstractmethod
    def get_universal_config_class(self) -> Optional[Type]:
        return None

    @property
    def project_config_cls(self) -> Type:
        klass = self.get_project_config_class()
        return klass or self.project_config_class

    @abstractmethod
    def get_project_config_class(self) -> Optional[Type]:
        return None

    @property
    def keychain_cls(self) -> Type:
        klass = self.get_keychain_class()
        return klass or self.keychain_class

    @abstractmethod
    def get_keychain_class(self) -> Optional[Type]:
        return None

    @property
    def keychain_key(self):
        return self.get_keychain_key()

    @abstractmethod
    def get_keychain_key(self):
        return None

    def _add_repo_to_path(self):
        if self.project_config and self.project_config.repo_root:
            sys.path.append(self.project_config.repo_root)

    def _load_universal_config(self):
        self.universal_config = self.universal_config_cls()

    def _load_project_config(self, *args, **kwargs):
        self.project_config = self.project_config_cls(
            self.universal_config, *args, **kwargs
        )
        if self.project_config is not None:
            self.project_config._add_tasks_directory_to_python_path()

    def _load_keychain(self):
        if self.keychain is not None:
            return

        keychain_key = self.keychain_key if self.keychain_cls.encrypted else None

        if self.project_config is None:
            self.keychain = self.keychain_cls(self.universal_config, keychain_key)
        else:
            self.keychain = self.keychain_cls(self.project_config, keychain_key)
            self.project_config.keychain = self.keychain

    def get_flow(self, name: str, options: Optional[dict] = None) -> FlowCoordinator:
        """Get a primed and ready-to-go flow coordinator."""
        if not self.project_config:
            raise ProjectConfigNotFound
        flow_config = self.project_config.get_flow(name)
        callbacks = self.callback_class()
        coordinator = FlowCoordinator(
            flow_config.project_config,
            flow_config,
            name=flow_config.name,
            options=options,
            skip=None,
            callbacks=callbacks,
        )
        return coordinator



from typing import NamedTuple


def add_after_statements(mappings: dict):
    """Automatically add CCI after: statements to the lookups
    in a mapping file"""
    indexed_by_sobject = _index_by_sobject(mappings)

    for idx, (_, mapping) in enumerate(mappings.items()):
        for lookup in mapping.get("lookups", {}).values():
            target_table = lookup["table"]
            # SnowfakeryPersonAccounts: Add this back in when Snowfakery is integrated with this code.
            # PersonContacts are not real, so skip them
            if target_table == "PersonContact":  # pragma: no cover
                raise NotImplementedError(
                    "This code is not yet tested for use with Snowfakery and Person Accounts"
                )
                # continue
            if isinstance(target_table, list):
                target_mapping_index = max(
                    [indexed_by_sobject[table] for table in target_table],
                    key=lambda index: index.first_instance,
                )
            else:
                target_mapping_index = indexed_by_sobject[target_table]

            if target_mapping_index.first_instance >= idx:
                if not lookup.get("after"):
                    lookup["after"] = target_mapping_index.last_step_name


class MappingIndex(NamedTuple):  # info needed by the algorithm above
    first_instance: int  # where was the first time this sobj was referenced?
    last_step_name: str  # where was the last (so far)?


def _index_by_sobject(mappings):
    """Figure out the order in which sobjects are loaded and assign an index to each sobject

    We can use this to detect sobjects out-of-order which require 'after' declarations."""
    indexed_by_sobject = {}
    for idx, (mapping_name, mapping) in enumerate(mappings.items()):
        # make an index of the order of objects
        sobject = mapping["sf_object"]
        existing_index = indexed_by_sobject.get(sobject)

        if existing_index:
            new_mi = MappingIndex(existing_index.first_instance, mapping_name)
        else:
            new_mi = MappingIndex(idx, mapping_name)
        indexed_by_sobject[sobject] = new_mi

    return indexed_by_sobject



import base64
import io
import zipfile

from cumulusci.core.config import BaseProjectConfig, UniversalConfig
from cumulusci.tasks.salesforce import CreatePackage

from .util import create_task


class TestCreatePackage:
    def test_get_package_zip(self):
        project_config = BaseProjectConfig(
            UniversalConfig(),
            {"project": {"package": {"name": "TestPackage", "api_version": "43.0"}}},
        )
        task = create_task(CreatePackage, project_config=project_config)
        package_zip = task._get_package_zip()
        zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(package_zip)), "r")
        package_xml = zf.read("package.xml")
        assert b"<fullName>TestPackage</fullName>" in package_xml
        zf.close()
